'use strict';

var dgram = require('dgram'),

    MULTICAST_PORT = 1900,
    MULTICAST_ADDR = "239.255.255.250",
    BROADCAST_SEND_PORT = 7070,
    BROADCAST_LSTN_PORT = 7071,
    BROADCAST_ADDR = "255.255.255.255",

    multicastSocket,
    broadcastSendSocket;

/**
 * Sends the SSDP and UDP search out
 * @param successCallback is called for every found Miniserver
 * @param errorCallback is called on any error
 */
exports.startSearch = function startLxFinder(successCallback, errorCallback) {
    this._performMulticastSearch(successCallback, errorCallback);
    this._performBroadcastSearch(successCallback, errorCallback);
};

/**
 * Stops both SSDP and UDP search immediately
 * @param [successCallback]
 * @param [errorCallback]
 */
exports.stopSearch = function stopLxFinder(successCallback, errorCallback) {
    try {
        multicastSocket && multicastSocket.close();
        broadcastSendSocket && broadcastSendSocket.close();
    } catch (e) {
        // Don't handle this error. There may be an error, if the socket hasn't been bound to the port yet.
        // This happens when the search is started and stopped nearly immediately
    }
    // Always call the success callback if available, there is no error to handle
    successCallback && successCallback();
};

/**
 * Sends out a SSDP (Multicast) search package and listens for its response
 * @param successCallback is called for every found Miniserver
 * @param errorCallback is called on any error
 * @private
 */
exports._performMulticastSearch = function _performMulticastSearch(successCallback, errorCallback) {
    try {
        multicastSocket = dgram.createSocket('udp4');
        multicastSocket.bind(function() {
            multicastSocket.setBroadcast(true);
        });
        multicastSocket.on('listening', function () {
            var message = new Buffer.from(
                "M-SEARCH LoxLIVE HTTP/1.1\r\n" +
                "Host:" + MULTICAST_ADDR + ":" + MULTICAST_PORT + "\r\n" +
                "Man:ssdp:discover\r\n" +
                "MX:3\r\n" +
                "ST:ssdp:all\r\n" +
                "\r\n");

            multicastSocket.on('message', function (message, rinfo) {
                var responseString = message.toString(),
                    nameRegex = /SERVER: ?Loxone Miniserver (.*) UPnP\/1.0/g,
                    addressRegex = /LOCATION: ?http:\/\/(.*)\/upnp.xml/g,
                    nameRegexRes = nameRegex.exec(responseString) || [],
                    addressRegexRes = addressRegex.exec(responseString) || [],
                    msName = nameRegexRes[1],
                    msAddress = addressRegexRes[1],
                    resObj = {};

                if (msAddress && msName) {
                    resObj[msAddress] = msName;
                    successCallback(resObj);
                }
            });

            multicastSocket.send(message, 0, message.length, MULTICAST_PORT, MULTICAST_ADDR);
        });
    } catch(e) {
        errorCallback(e)
    }
};

/**
 * Sends out a UDP (Broadcast) search package and listens for its response
 * @param successCallback is called for every found Miniserver
 * @param errorCallback is called on any error
 * @private
 */
exports._performBroadcastSearch = function _performBroadcastSearch(successCallback, errorCallback) {
    try {
        var msgToSend = new Buffer.from([0x00]),
            retransmitTimeout = null;
        broadcastSendSocket = dgram.createSocket('udp4');
        broadcastSendSocket.bind(BROADCAST_LSTN_PORT, function() {
            broadcastSendSocket.setBroadcast(true);
        });
        broadcastSendSocket.on('listening', function () {
            broadcastSendSocket.on('message', function (message, rinfo) {
                var responseString = message.toString(),
                    nameRegex = RegExp("LoxLIVE: ?(.*) " + rinfo.address, "g"),
                    portRegex = RegExp(" " + rinfo.address + ":(\\d*) ", "g"),
                    nameRegexRes = nameRegex.exec(responseString) || [],
                    portRegexRes = portRegex.exec(responseString) || [],
                    msName = nameRegexRes[1],
                    msAddress = rinfo.address + ":" + portRegexRes[1],
                    resObj = {};

                if (msAddress && msName) {
                    resObj[msAddress] = msName;
                    successCallback(resObj);
                }
            });

            broadcastSendSocket.send(msgToSend, 0, msgToSend.length, BROADCAST_SEND_PORT, BROADCAST_ADDR);
            // Sometimes Miniservers won't respond on the first broadcast, send another one...
            retransmitTimeout = setTimeout(function() {
                broadcastSendSocket.send(msgToSend, 0, msgToSend.length, BROADCAST_SEND_PORT, BROADCAST_ADDR);
                retransmitTimeout = null;
            }, 50);
        });

        // Prevent any retransmit after the socket has been closed
        broadcastSendSocket.on("close", function() {
           if (retransmitTimeout !== null) {
               clearTimeout(retransmitTimeout);
               retransmitTimeout = null;
           }
        });
    } catch (e) {
        errorCallback(e)
    }
};
