const childProcess = require("child_process"),
    path = require("path"),
    fs = require("fs"),
    fsExtra = require("fs-extra"),
    glob = require("glob"),
    Q = require("q"),
    Listr = require('listr'),
    PJSIPCloner = require("./clonePJSIP.js"),
    { pjsip, name } = require('../package.json'),
    PLATFORMS = [
        "darwin",
        "win32"
    ],
    PLATFORM_EMOJI_MAP = {
        darwin: "",
        win32: "⊞",
        linux: "🐧"
    },
    PLATFORM_NODE_GYP_MAP = {
        darwin: "mac",
        win32: "win",
        linux: "linux"
    },
    /**
     * We support these architectures
     * @type {{win32: string[], darwin: string[]}}
     */
    PLATFORM_ARCH_MAP = {
        darwin: [
            "x64",
            "arm64"
        ],
        win32: [
            "x64"
        ]
    },
    DARWIN_ARCH_X_COMPILE_HOST_MAP = {
        arm64: "arm-apple-darwin",
        x64: "x86_64-apple-darwin"
    },
    DARWIN_ARCH_X_COMPILE_MAP = {
        arm64: "arm64",
        x64: "x86_64"
    }
    DEFINE_MAP = {
        /*darwin_arm64: [
            "ARM",
            "PJ_IS_LITTLE_ENDIAN=1",
            "PJ_IS_BIG_ENDIAN=0"
        ],
        darwin_x64: [
            "x86"
        ]*/
    };

return new Listr([
    {
        title: `Compile ${name}`,
        task: () => {
            return new Listr([
                ...PJSIPCloner.tasks,
                {
                    title: `🏗 Building ${PLATFORMS.join(",")}`,
                    task: () => new Listr(PLATFORMS.map(getTaskForPlatform), { concurrent: true })
                }
            ])
        }
    }
]).run();

function supportsPlatformBuild(nodePlatform) {
    if (process.platform === nodePlatform) {
        return true;
    } else if (nodePlatform === "win32" && process.platform === "darwin") {
        return require("parallels-helper").getCompatibleVm()
    } else {
        return false;
    }
}

function getTaskForPlatform(nodePlatform) {
    return {
        title: `${PLATFORM_EMOJI_MAP[nodePlatform]}`,
        enabled: () => supportsPlatformBuild(nodePlatform),
        task: () => {
            let platformPath = path.join(__dirname, "../native/pjsip/platforms", PLATFORM_NODE_GYP_MAP[nodePlatform]),
                platformSrcPath = path.join(platformPath, "src");

            return new Listr([
                {
                    title: "🧹 Cleaning up",
                    enabled: () => fs.existsSync(platformSrcPath),
                    task: () => {
                        let def = Q.defer();
                        fs.rm(platformSrcPath, { recursive: true }, error => {
                            if (error) {
                                def.reject(error);
                            } else {
                                def.resolve();
                            }
                        });
                        return def.promise;
                    }
                },
                {
                    title: `📂 Copy PJSIP ${pjsip.tag}`,
                    task: () => {
                        let def = Q.defer();
                        fsExtra.copy(PJSIPCloner.DIR, platformSrcPath, error => {
                            if (error) {
                                def.reject(error);
                            } else {
                                fs.rm(path.join(platformSrcPath, ".git"), { recursive: true }, error => {
                                    if (error) {
                                        def.reject(error);
                                    } else {
                                        def.resolve();
                                    }
                                });
                            }
                        });
                        return def.promise;
                    }
                },
                {
                    title: '📃 Copy "config_site.h"',
                    task: () => {
                        let def = Q.defer();
                        fsExtra.copy(
                            path.join(platformPath, "config_site.h"),
                            path.join(platformPath, path.join("src", "pjlib", "include", "pj", "config_site.h")),
                            error => {
                                if (error) {
                                    def.reject(error);
                                } else {
                                    def.resolve();
                                }
                            }
                        );
                        return def.promise;
                    }
                },
                {
                    title: "🏗 PJSIP libs",
                    task: () => {
                        return new Listr(PLATFORM_ARCH_MAP[nodePlatform].map(arch => {
                            return getBuildPJSIPLibTask({
                                platform: nodePlatform,
                                arch: arch,
                                platformSrcPath: platformSrcPath
                            })
                        }, {
                            concurrent: true
                        }));
                    }
                }
            ]);
        }
    }
}

function getBuildPJSIPLibTask({ platform, arch, platformSrcPath}) {
    let resPath = path.resolve(platformSrcPath, "..", "res", pjsip.tag, arch);
    return {
        title: `🏗 ${PLATFORM_EMOJI_MAP[platform]} ${arch}`,
        task: () => {
            return new Listr([
                {
                    title: "Preparing Compiling env",
                    task: () => {
                        let def = Q.defer(),
                            newPlatformSrcPath = `${platformSrcPath}_${arch}`;
                        fsExtra.copy(platformSrcPath, newPlatformSrcPath, error => {
                            if (error) {
                                def.reject(error);
                            } else {
                                platformSrcPath = newPlatformSrcPath;
                                def.resolve();
                            }
                        });
                        return def.promise;
                    }
                },
                {
                    title: "🖊 Writing defines",
                    enabled: () => `${platform}_${arch}` in DEFINE_MAP,
                    task: () => {
                        let configPath = path.join(platformSrcPath, "pjlib", "include", "pj", "config_site.h"),
                            config = fs.readFileSync(configPath, "utf8"),
                            defines = DEFINE_MAP[`${platform}_${arch}`];

                        defines.forEach(define => {
                            config += `\n#define ${define.replace("=", " ")}`;
                        });

                        fs.writeFileSync(configPath, config, "utf8");
                    }
                },
                {
                    title: "🏗 PJSIP",
                    enabled: () => !fs.existsSync(path.resolve(resPath, "libs")),
                    task: () => {
                        let def = Q.defer(),
                            msBuildPath = "C:\\Program Files (x86)\\Microsoft Visual Studio\\2017\\BuildTools\\MSBuild\\15.0\\Bin\\MSBuild.exe",
                            slnPath = path.resolve(platformSrcPath, "pjproject-vs14.sln");

                        switch (process.platform) {
                            case "darwin":
                                // We build macOS on macOS
                                if (platform === "darwin") {
                                    childProcess.exec(`CFLAGS="-arch ${DARWIN_ARCH_X_COMPILE_MAP[arch]}" LDFLAGS="-arch ${DARWIN_ARCH_X_COMPILE_MAP[arch]}" ./configure --host=${DARWIN_ARCH_X_COMPILE_HOST_MAP[arch]} && make dep && make clean && make`, { cwd: platformSrcPath }, (error, stdout) => {
                                        if (error) {
                                            def.reject(error);
                                        } else {
                                            def.resolve(stdout);
                                        }
                                    });
                                // We build windows on macOS
                                } else if (platform === "win32") {
                                    def.resolve(
                                        require("parallels-helper").getCompatibleVm().then(vm => {
                                            return require("parallels-helper").exec([
                                                `"${msBuildPath}"`,
                                                `"${require("parallels-helper").windowsifyPath(slnPath).replace(/\\/g, "\\\\")}"`,
                                                "/target:pjsua",
                                                `/property:Platform=${arch}`,
                                                "/property:Configuration=Release-Static",
                                                "-m"
                                            ].join(" "), vm);
                                        })
                                    )
                                } else {
                                    def.reject(new Error(`Unsupported platform ${platform}`));
                                }
                                break;
                            case "win32":
                                if (platform === "win32") {
                                    childProcess.exec([
                                        `"${msBuildPath}"`,
                                        `"${slnPath}"`,
                                        "/target:pjsua",
                                        `/property:Platform=${arch}`,
                                        "/property:Configuration=Release-Static",
                                        "-m"
                                    ].join(" "), { cwd: platformSrcPath }, (error, stdout) => {
                                        if (error) {
                                            def.reject(error);
                                        } else {
                                            def.resolve(stdout);
                                        }
                                    });
                                } else {
                                    def.reject(new Error(`Unsupported platform ${platform}`));
                                }
                                break;
                            default:
                                def.reject(new Error(`Unsupported host platform ${process.platform}`));
                        }
                        return def.promise;
                    }
                },
                {
                    title: "💾 Save PJSIP lib(s)",
                    enabled: () => !fs.existsSync(path.resolve(resPath, "libs")),
                    task: () => {
                        let def = Q.defer(),
                            globPattern;
                        switch (platform) {
                            case "darwin":
                                globPattern = path.join(platformSrcPath, "**/lib/*.a");
                                break;
                            case "win32":
                                globPattern = path.join(platformSrcPath, "**/lib/*.lib");
                                break;
                        }

                        globPattern && glob(globPattern, { absolute: true }, (err, files) => {
                            if (err) {
                                def.reject(err);
                            } else {
                                def.resolve(files);
                            }
                        });

                        return def.promise.then(files => {
                            return Q.all(files.map(file => {
                                return fsExtra.copy(
                                    file,
                                    path.join(resPath, "libs", path.basename(file))
                                );
                            }));
                        });
                    }
                },
                {
                    title: "💾 Save PJSIP includes",
                    enabled: () => !fs.existsSync(path.resolve(resPath, "includes")),
                    task: () => {
                        let def = Q.defer();

                        glob(path.join(platformSrcPath, "**/include"), { absolute: true }, (err, files) => {
                            if (err) {
                                def.reject(err);
                            } else {
                                def.resolve(files);
                            }
                        });

                        return def.promise.then(files => {
                            return Q.all(files.map(file => {
                                return fsExtra.copy(
                                    file,
                                    path.join(resPath, "includes", file.replace(platformSrcPath, ""))
                                );
                            }));
                        });
                    }
                }/*,
                {
                    title: "🏗 nodesip",
                    task: () => {
                        return new Listr([
                            {
                                title: "🏗 node",
                                task: () => {
                                    let def = Q.defer(),
                                        buildCmd = [
                                            "npx",
                                            "node-pre-gyp",
                                            "rebuild",
                                            // That is the arch for the node-gyp binding file
                                            `--target_arch=${arch}`,
                                            //`--target_platform=${platform}`,
                                            "--verbose"
                                        ].join(" ")
                                    childProcess.exec(buildCmd, { cwd: path.join(__dirname, "..") }, (error, stdout) => {
                                        if (error) {
                                            def.reject(error);
                                            fs.writeFileSync( path.join(__dirname, "..", "node-error.log"), stdout, "utf8");
                                        } else {
                                            def.resolve(stdout);
                                        }
                                    });
                                    return def.promise;
                                }
                            },
                            {
                                title: "🏗 electron",
                                task: () => {
                                    let def = Q.defer(),
                                        buildCmd = [
                                            "npx",
                                            "electron-rebuild",
                                            "-f",
                                            `--arch=${arch}`,
                                            // That is the arch for the node-gyp binding file
                                            `--target_arch=${arch}`,
                                            //`--target_platform=${platform}`,
                                            "--verbose"
                                        ].join(" ");
                                    childProcess.exec(buildCmd, { cwd: path.join(__dirname, "..") }, (error, stdout) => {
                                        if (error) {
                                            def.reject(error);
                                            fs.writeFileSync( path.join(__dirname, "..", "electron-error.log"), stdout, "utf8");
                                        } else {
                                            def.resolve(stdout);
                                        }
                                    });
                                    return def.promise;
                                }
                            }
                        ])
                    }
                }*/,
                {
                    title: "🧹 Compiling env",
                    task: () => {
                        let def = Q.defer();
                        fs.rm(platformSrcPath, { recursive: true }, error => {
                            if (error) {
                                def.reject(error);
                            } else {
                                def.resolve();
                            }
                        });
                        return def.promise;
                    }
                }
            ]);
        }
    }
}